@@ -10,7 +10,7 @@ This extension helps manage:
 This code is public domain.
 """
 
-from typing import Optional, Callable, Dict, List, Set, Any
+from typing import Optional, Callable, Dict, List, Set, Any, Union
 
 import functools
 import inspect
@@ -28,19 +28,17 @@ from flask import session, jsonify, Blueprint, make_response, abort, \
     redirect, url_for, after_this_request, send_file, send_from_directory, \
     safe_join, escape, Markup, render_template, current_app, g
 
-from werkzeug.datastructures import CombinedMultiDict, MultiDict
-
 import logging
 log = logging.getLogger("fsa")
 
 # get module version
-__version__ = "4.5.0"
+__version__ = "4.6.1"
 # FIXME currently broken because of dependency issues with typing_extensions
 # import pkg_resources as pkg  # type: ignore
 # __version__ = pkg.require("FlaskSimpleAuth")[0].version
 
 
-class AuthException(BaseException):
+class FSAException(BaseException):
     """Exception class to carry fields for an error Response."""
 
     def __init__(self, message: str, status: int):
@@ -196,33 +194,6 @@ class Reference:
         return self._obj.__gt__(o)
 
 
-#
-# TODO
-# - LRU? LFU?
-# - automatic reset based on cache efficiency? expansion?
-#
-class CacheOK:
-    """Positive caching decorator for boolean functions.
-
-    Cache True answers, but still forwards False answers to the underlying
-    function.
-    """
-
-    def __init__(self, fun: Callable[[List[Any]], bool]):
-        self.__wrapped__ = fun
-        self._cache: Set[Any] = set()
-        self.cache_clear = self._cache.clear
-
-    def __call__(self, *args):
-        if args in self._cache:
-            return True
-        else:
-            ok = self.__wrapped__(*args)
-            if ok:
-                self._cache.add(args)
-            return ok
-
-
 class Flask(flask.Flask):
     """Flask class wrapper.
 
@@ -303,18 +274,20 @@ class Flask(flask.Flask):
 
 # all possible directives
 _DIRECTIVES = {
-    "FSA_401_REDIRECT", "FSA_AUTH", "FSA_CACHE_SIZE", "FSA_CHECK",
+    "FSA_401_REDIRECT", "FSA_AUTH", "FSA_CHECK",
+    "FSA_CACHE", "FSA_CACHE_SIZE", "FSA_CACHE_OPTS",
     "FSA_FAKE_LOGIN", "FSA_GET_USER_PASS", "FSA_HTTP_AUTH_OPTS",
-    "FSA_MODE", "FSA_PARAM_PASS", "FSA_PARAM_USER", "FSA_PASSWORD_OPTIONS",
+    "FSA_MODE", "FSA_PARAM_PASS", "FSA_PARAM_USER", "FSA_PASSWORD_OPTS",
     "FSA_PASSWORD_SCHEME", "FSA_SKIP_PATH", "FSA_TOKEN_ALGO",
     "FSA_TOKEN_CARRIER", "FSA_TOKEN_DELAY", "FSA_TOKEN_GRACE",
     "FSA_TOKEN_LENGTH", "FSA_TOKEN_NAME", "FSA_REALM",
     "FSA_TOKEN_SECRET", "FSA_TOKEN_SIGN", "FSA_TOKEN_TYPE",
     "FSA_TOKEN_RENEWAL", "FSA_URL_NAME", "FSA_USER_IN_GROUP",
-    "FSA_LOGGING_LEVEL", "FSA_CORS", "FSA_CORS_OPTIONS",
+    "FSA_LOGGING_LEVEL", "FSA_CORS", "FSA_CORS_OPTS",
     "FSA_PASSWORD_LEN", "FSA_PASSWORD_RE",
 }
 
+_DEFAULT_CACHE_SIZE = 16384
 
 # actual extension
 class FlaskSimpleAuth:
@@ -323,13 +296,13 @@ class FlaskSimpleAuth:
     def __init__(self, app: flask.Flask = None):
         """Constructor parameter: flask application to extend."""
         self._app = app
-        self._maxsize = 1024
         self._get_user_pass = None
         self._user_in_group = None
         self._auth: List[str] = []
         self._saved_auth: Optional[List[str]] = None
         self._http_auth = None
         self._pm = None
+        self._cache: Optional[Callable[[Any], Callable]] = None
         # actual main initialization is deferred
         self._initialized = False
 
@@ -353,7 +326,7 @@ class FlaskSimpleAuth:
                 return
         try:
             self.get_user()
-        except AuthException as e:
+        except FSAException as e:
             return self._Resp(e.message, e.status)
 
     def _auth_after_cleanup(self, res: Response):
@@ -391,7 +364,7 @@ class FlaskSimpleAuth:
                 if self._name in request.cookies:
                     user, exp = self._get_any_token_auth_exp(request.cookies[self._name])
                     # renew token when closing expiration
-                    limit = dt.datetime.utcnow() + self._renewal * dt.timedelta(minutes=self._delay)
+                    limit = dt.datetime.now(dt.timezone.utc) + self._renewal * dt.timedelta(minutes=self._delay)
                     set_cookie = exp < limit
                 else:
                     set_cookie = True
@@ -432,8 +405,8 @@ class FlaskSimpleAuth:
         # get the actual function when regenerating caches
         while hasattr(fun, "__wrapped__"):
             fun = fun.__wrapped__
-        return fun if not fun or self._maxsize == 0 else \
-            functools.lru_cache(maxsize=self._maxsize)(fun)
+        return fun if not fun or self._cache is None else \
+            self._cache(**self._cache_opts)(fun)
 
     def _params(self):
         """Get request parameters wherever they are."""
@@ -443,6 +416,7 @@ class FlaskSimpleAuth:
             # reimplement "request.values" after Flask 2.0 regression
             # https://github.com/pallets/werkzeug/pull/2037
             # https://github.com/pallets/flask/issues/4120
+            from werkzeug.datastructures import CombinedMultiDict, MultiDict
             return CombinedMultiDict([MultiDict(d) if not isinstance(d, MultiDict) else d
                                       for d in (request.args, request.form)])
 
@@ -503,19 +477,49 @@ class FlaskSimpleAuth:
         self._mode = conf.get("FSA_MODE", "lazy")
         assert self._mode in ("always", "lazy", "all")
         self._check: bool = conf.get("FSA_CHECK", True)
-        # for web apps…
+        self._skip_path = [re.compile(r).match for r in conf.get("FSA_SKIP_PATH", [])]
+        #
+        # web apps…
+        #
         self._cors: bool = conf.get("FSA_CORS", False)
-        self._cors_options: Dict[str, Any] = conf.get("FSA_CORS_OPTIONS", {})
+        self._cors_opts: Dict[str, Any] = conf.get("FSA_CORS_OPTS", {})
         if self._cors:
             from flask_cors import CORS  # type: ignore
-            CORS(self._app, **self._cors_options)
-        self._maxsize = conf.get("FSA_CACHE_SIZE", 1024)
-        import re
-        self._skip_path = [re.compile(r).match for r in conf.get("FSA_SKIP_PATH", [])]
-        # for web apps
+            CORS(self._app, **self._cors_opts)
         self._401_redirect = conf.get("FSA_401_REDIRECT", None)
         self._url_name = conf.get("FSA_URL_NAME", "URL" if self._401_redirect else None)
         #
+        # internal cache management for passwords, permissions, tokens…
+        #
+        self._cache_opts: Dict[str, Any] = conf.get("FSA_CACHE_OPTS", {})
+        if "FSA_CACHE_SIZE" in conf or "maxsize" not in self._cache_opts:
+            self._cache_opts.update(maxsize=conf.get("FSA_CACHE_SIZE", _DEFAULT_CACHE_SIZE))
+        cache: Union[str, Callable[[Any], Callable]] = conf.get("FSA_CACHE", "fc-lru")
+        if isinstance(cache, str):
+            if cache == "ttl":
+                import cachetools.func
+                self._cache = cachetools.func.ttl_cache
+                if "ttl" not in self._cache_opts:
+                    self._cache_opts.update(ttl=60 * 10)  # 10 minutes default
+            elif cache == "lru":
+                import cachetools.func
+                self._cache = cachetools.func.lru_cache
+            elif cache == "lfu":
+                import cachetools.func
+                self._cache = cachetools.func.lfu_cache
+            elif cache == "fifo":
+                import cachetools.func
+                self._cache = cachetools.func.fifo_cache
+            elif cache == "rr":
+                import cachetools.func
+                self._cache = cachetools.func.rr_cache
+            elif cache == "fc-lru":
+                self._cache = functools.lru_cache
+            else:
+                raise Exception(f"Unexpected FSA_CACHE: {cache}")
+        else:  # hopefully a callable
+            self._cache = cache
+        #
         # token setup
         #
         self._token = conf.get("FSA_TOKEN_TYPE", "fsa")
@@ -686,7 +690,7 @@ class FlaskSimpleAuth:
                 log.warning("plaintext password manager is a bad idea")
             # passlib context is a pain, you have to know the scheme name to set its
             # round. Ident "2y" is same as "2b" but apache compatible.
-            options = conf.get("FSA_PASSWORD_OPTIONS",
+            options = conf.get("FSA_PASSWORD_OPTS",
                                {"bcrypt__default_rounds": 4,
                                 "bcrypt__default_ident": "2y"})
             from passlib.context import CryptContext  # type: ignore
@@ -717,14 +721,14 @@ class FlaskSimpleAuth:
         user = params.get(self._login, None)
         # it could check that the user exists in db
         if not user:
-            raise AuthException("missing login parameter", 401)
+            raise FSAException("missing login parameter", 401)
         return user
 
     #
     # PASSWORD MANAGEMENT
     #
     # FSA_PASSWORD_SCHEME: name of password scheme for passlib context
-    # FSA_PASSWORD_OPTIONS: further options for passlib context
+    # FSA_PASSWORD_OPTS: further options for passlib context
     # FSA_PASSWORD_LEN: minimal length of provided passwords
     # FSA_PASSWORD_RE: list of re a password must match
     #
@@ -740,10 +744,10 @@ class FlaskSimpleAuth:
         # check password quality
         if check:
             if len(pwd) < self._password_len:
-                raise AuthException(f"password is too short, must be at least {self._password_len}", 400)
+                raise FSAException(f"password is too short, must be at least {self._password_len}", 400)
             for r in self._password_re:
                 if not re.search(r, pwd):
-                    raise AuthException(f"password must match {r}", 400)
+                    raise FSAException(f"password must match {r}", 400)
         return self._pm.hash(pwd)
 
     def _check_password(self, user, pwd):
@@ -755,12 +759,12 @@ class FlaskSimpleAuth:
         ref = self._get_user_pass(user)
         if not ref:
             log.debug(f"AUTH (password): no such user ({user})")
-            raise AuthException(f"no such user: {user}", 401)
+            raise FSAException(f"no such user: {user}", 401)
         if not isinstance(ref, (str, bytes)):
-            raise AuthException("get_user_pass must return None or str or bytes", 500)
+            raise FSAException("get_user_pass must return None or str or bytes", 500)
         if not self.check_password(pwd, ref):
             log.debug(f"AUTH (password): invalid password for {user}")
-            raise AuthException(f"invalid password for {user}", 401)
+            raise FSAException(f"invalid password for {user}", 401)
         return user
 
     #
@@ -779,11 +783,11 @@ class FlaskSimpleAuth:
             user = self._http_auth.authenticate(auth, password)
             if user is not None and user is not False:
                 return auth.username if user is True else user
-        except AuthException as error:
+        except FSAException as error:
             log.debug(f"AUTH (http-*): bad authentication {error}")
             raise error
         log.debug("AUTH (http-*): bad authentication")
-        raise AuthException("failed HTTP authentication", 401)
+        raise FSAException("failed HTTP authentication", 401)
 
     #
     # HTTP BASIC AUTH
@@ -796,16 +800,16 @@ class FlaskSimpleAuth:
         log.debug(f"auth: {auth}")
         if not auth:
             log.debug("AUTH (basic): missing authorization header")
-            raise AuthException("missing authorization header", 401)
+            raise FSAException("missing authorization header", 401)
         if auth[:6] != "Basic ":
             log.debug(f"AUTH (basic): unexpected auth \"{auth}\"")
-            raise AuthException("unexpected authorization header", 401)
+            raise FSAException("unexpected authorization header", 401)
         try:
             user, pwd = b64.b64decode(auth[6:]).decode().split(":", 1)
             self._check_password(user, pwd)
         except Exception as e:
             log.debug(f"AUTH (basic): error while decoding auth \"{auth}\" ({e})")
-            raise AuthException("decoding error on authorization header", 401)
+            raise FSAException("decoding error on authorization header", 401)
         return user
 
     #
@@ -822,10 +826,10 @@ class FlaskSimpleAuth:
         params = self._params()
         user = params.get(self._userp, None)
         if not user:
-            raise AuthException(f"missing login parameter: {self._userp}", 401)
+            raise FSAException(f"missing login parameter: {self._userp}", 401)
         pwd = params.get(self._passp, None)
         if not pwd:
-            raise AuthException(f"missing password parameter: {self._passp}", 401)
+            raise FSAException(f"missing password parameter: {self._passp}", 401)
         self._check_password(user, pwd)
         return user
 
@@ -836,7 +840,7 @@ class FlaskSimpleAuth:
         """Get user from basic or param authentication."""
         try:
             return self._get_basic_auth()
-        except AuthException:  # failed, let's try param
+        except FSAException:  # failed, let's try param
             return self._get_param_auth()
 
     #
@@ -879,11 +883,11 @@ class FlaskSimpleAuth:
         p = re.match(r"^(\d{4})(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)$", ts)
         if not p:
             raise Exception(f"unexpected timestamp format: {ts}")
-        return dt.datetime(*[int(p[i]) for i in range(1, 7)])
+        return dt.datetime(*[int(p[i]) for i in range(1, 7)], tzinfo=dt.timezone.utc)
 
     def _get_fsa_token(self, realm, user, delay, secret):
         """Compute a signed token for "user" valid for "delay" minutes."""
-        limit = self._to_timestamp(dt.datetime.utcnow() + dt.timedelta(minutes=delay))
+        limit = self._to_timestamp(dt.datetime.now(dt.timezone.utc) + dt.timedelta(minutes=delay))
         data = f"{realm}:{user}:{limit}"
         sig = self._cmp_sig(data, secret)
         return f"{data}:{sig}"
@@ -896,7 +900,7 @@ class FlaskSimpleAuth:
         - iss: issuer (not used)
         - aud = audience (the realm)
         """
-        exp = dt.datetime.utcnow() + dt.timedelta(minutes=delay)
+        exp = dt.datetime.now(dt.timezone.utc) + dt.timedelta(minutes=delay)
         import jwt
         return jwt.encode({"exp": exp, "sub": user, "aud": realm},
                           secret, algorithm=self._algo)
@@ -928,17 +932,17 @@ class FlaskSimpleAuth:
         # check realm
         if realm != self._realm:
             log.debug(f"AUTH (fsa token): unexpected realm {realm}")
-            raise AuthException(f"unexpected realm: {realm}", 401)
+            raise FSAException(f"unexpected realm: {realm}", 401)
         # check signature
         ref = self._cmp_sig(f"{realm}:{user}:{slimit}", self._secret)
         if ref != sig:
             log.debug("AUTH (fsa token): invalid signature")
-            raise AuthException("invalid fsa auth token signature", 401)
+            raise FSAException("invalid fsa auth token signature", 401)
         # check limit with a grace time
-        now = dt.datetime.utcnow() - dt.timedelta(minutes=self._grace)
+        now = dt.datetime.now(dt.timezone.utc) - dt.timedelta(minutes=self._grace)
         if now > limit:
             log.debug("AUTH (fsa token): token {token} has expired")
-            raise AuthException("expired fsa auth token", 401)
+            raise FSAException("expired fsa auth token", 401)
         # all is well
         return user, limit
 
@@ -952,19 +956,19 @@ class FlaskSimpleAuth:
         try:
             data = jwt.decode(token, self._secret, leeway=self._grace * 60.0,
                               audience=self._realm, algorithms=[self._algo])
-            exp = dt.datetime.fromtimestamp(data["exp"])
+            exp = dt.datetime.fromtimestamp(data["exp"], tz=dt.timezone.utc)
             return data["sub"], exp
         except jwt.ExpiredSignatureError:
             log.debug(f"AUTH (jwt token): token {token} has expired")
-            raise AuthException("expired jwt auth token", 401)
+            raise FSAException("expired jwt auth token", 401)
         except Exception as e:
             log.debug(f"AUTH (jwt token): invalid token ({e})")
-            raise AuthException("invalid jwt token", 401)
+            raise FSAException("invalid jwt token", 401)
 
     def _get_any_token_auth_exp(self, token):
         """return validated user and expiration."""
         if not token:
-            raise AuthException("missing token", 401)
+            raise FSAException("missing token", 401)
         return \
             self._get_fsa_token_auth(token) if self._token == "fsa" else \
             self._get_jwt_token_auth(token)
@@ -973,10 +977,10 @@ class FlaskSimpleAuth:
         """Tell whether token is ok: return validated user or None."""
         user, exp = self._get_any_token_auth_exp(token)
         # recheck token expiration
-        now = dt.datetime.utcnow() - dt.timedelta(minutes=self._grace)
+        now = dt.datetime.now(dt.timezone.utc) - dt.timedelta(minutes=self._grace)
         if now > exp:
             log.debug(f"AUTH (token): token {token} has expired")
-            raise AuthException("expired auth token", 401)
+            raise FSAException("expired auth token", 401)
         return user
 
     def _get_token_auth(self) -> Optional[str]:
@@ -1036,7 +1040,7 @@ class FlaskSimpleAuth:
                 self._user = self._FSA_AUTH[a](self)
                 if self._user:
                     break
-            except AuthException as e:
+            except FSAException as e:
                 lae = e
             # FIXME other exceptions?
 
@@ -1045,7 +1049,7 @@ class FlaskSimpleAuth:
 
         # rethrow last auth exception on failure
         if required and not self._user:
-            raise lae or AuthException("missing authentication", 401)
+            raise lae or FSAException("missing authentication", 401)
 
         log.debug(f"get_user({self._auth}): {self._user}")
         return self._user
@@ -1102,7 +1106,10 @@ class FlaskSimpleAuth:
                 if NONE in groups or None in groups:
                     return self._Resp("", 403)
                 if ANY in groups:
-                    return fun(*args, **kwargs)
+                    try:
+                        return fun(*args, **kwargs)
+                    except FSAException as e:
+                        return self._Resp(e.message, e.status)
                 # get user if needed
                 if not self._user:
                     # no current user, try to get one?
@@ -1115,22 +1122,28 @@ class FlaskSimpleAuth:
                             self._saved_auth, self._auth = self._auth, auth
                         try:
                             self._user = self.get_user()
-                        except AuthException as ae:
-                            return self._Resp(ae.message, ae.status)
+                        except FSAException as e:
+                            return self._Resp(e.message, e.status)
                     else:
                         return self._Resp("", 401)
                 if not self._user:  # pragma: no cover
                     return self._Resp("", 401)  # should be unreachable
                 # shortcut for authenticated users
                 if ALL in groups:
-                    return fun(*args, **kwargs)
+                    try:
+                        return fun(*args, **kwargs)
+                    except FSAException as e:
+                        return self._Resp(e.message, e.status)
                 # check against all authorized groups/roles
                 for r in groups:
                     uig = self._user_in_group(self._user, r)
                     if not isinstance(uig, bool):
                         return self._Resp("user_in_group must return a boolean", 500)
                     if uig:
-                        return fun(*args, **kwargs)
+                        try:
+                            return fun(*args, **kwargs)
+                        except FSAException as e:
+                            return self._Resp(e.message, e.status)
                 # else no matching group
                 return self._Resp("", 403)
 
@@ -1220,7 +1233,10 @@ class FlaskSimpleAuth:
                             kwargs[p] = params[p]
 
                 # then call the initial function
-                return fun(*args, **kwargs)
+                try:
+                    return fun(*args, **kwargs)
+                except FSAException as e:
+                    return self._Resp(e.message, e.status)
 
             return wrapper
 
@@ -1275,7 +1291,7 @@ class FlaskSimpleAuth:
             log.warning(f"missing authorize on route \"{rule}\" makes it 403 Forbidden")
 
         def decorate(fun):
-            self.add_url_rule(rule, view_func=fun, **options)
+            return self.add_url_rule(rule, view_func=fun, **options)
         return decorate
 
     # support Flask 2.0 per-method decorator shortcuts
